#!/usr/bin/env python

"""
<doc:par>This module implements <application>make</application> functionality.</doc:par>

<doc:par>Like <application>make</application> it allows you to specify dependencies
between files and actions to be executed when files don't exist or are out of date
with respect to one of their sources. But unlike <application>make</application> you
can do this in a object oriented way and targets are not only limited to files,
but you can implement e.g. dependencies on database records by implementing new
<pyref class="Target">Target</pyref> classes.</doc:par>

<doc:par>Relevant classes are:
<ulist>
<item><pyref class="Project">Project</pyref>, which is the container for all targets,
actions and dependencies in a project,</item>
<item><pyref class="Target">Target</pyref> (and subclasses), which are files
(or other entities like database records) that have to be built by actions,</item>
<item><pyref class="Dep">Dep</pyref> (and subclasses) which specify types
of dependencies between targets and</item>
<item><pyref class="Action">Action</pyref> (and subclasses) which specify how targets
are (re)built if they don't exist or aren't up to date.</item>
</ulist>
</doc:par>
"""

import sys, os, os.path, time, getopt, types

from mx import DateTime

import ansistyle, fileutils

indent = " "*8

class Project:
	"""
	<doc:par>A <pyref class="Project" nolink>Project</pyref> collects all <pyref class="Target">Target</pyref>s
	from a project. It is responsible for initiating the build process and for generating a report about
	the progress of the build process.</doc:par>
	"""
	def __init__(self, verbose=1, ignoreExceptions=0, color=None):
		self.targets = {}
		self.rebuilt = 0
		self.failed = 0
		self.verbose = verbose
		self.ignoreExceptions = ignoreExceptions
		if color is None:
			color = os.getenv("MAKE_REPRANSI")
			if color is None:
				color = 0
			else:
				color = int(color)
				if not sys.platform.startswith("linux"):
					color = 0
		self.color = color
		self._initColors()

	def _initColors(self):
		"""
		initialize the color classes.
		"""
		self.TextTime = ansistyle.Text5
		self.TextSize = ansistyle.Text4
		self.TextID = ansistyle.Text2
		self.TextAction = ansistyle.Text15
		self.TextError = ansistyle.Text9

	def secondsStr(self, delta):
		if self.color:
			return ansistyle.Text(self.TextTime("%.03f" % delta.seconds), "s")
		else:
			return "%.03fs" % delta.seconds

	def sizeStr(self, bytes):
		if self.color:
			return ansistyle.Text(self.TextSize("%d" % bytes), " bytes")
		else:
			return "%d bytes" % bytes

	def actionStr(self, action):
		s = str(action)
		if self.color:
			s = self.TextAction(s)
		return s

	def errorStr(self, text):
		if self.color:
			return self.TextError(text)
		else:
			return text

	def idStr(self, id):
		s = str(id)
		if self.color:
			s = self.TextID(s)
		return s

	def __getitem__(self, name):
		return self.targets[name]

	def build(self, *targets):
		"""
		<doc:par>build all targets in <pyref arg="targets">targets</pyref>.</doc:par>
		
		<doc:par>Items in <pyref arg="targets">targets</pyref> must be <pyref class="Target">Target</pyref>
		instances or target ids. If an id can't be found, it will be wrapped in a
		<pyref module="fileutils" class="Filename">Filename</pyref> instance and retried.</doc:par>
		"""
		t1 = DateTime.now()
		self.rebuilt = 0
		self.failed = 0
		for target in targets:
			if not isinstance(target, Target):
				try:
					target = self[target]
				except KeyError:
					target = self[fileutils.Filename(target)] # retry with a filename
			target.update()
		t2 = DateTime.now()
		self.report("build %s\n%s%s (%d targets; %d rebuilt; %d failed)\n" % (self.__class__.__name__, indent, self.secondsStr(t2-t1), len(self.targets.keys()), self.rebuilt, self.failed))
		self.rebuilt = 0
		self.failed = 0

	def buildWithArgs(self, args):
		verbose = self.verbose
		ignoreExceptions = self.ignoreExceptions
		color = self.color
		(options, args) = getopt.getopt(args, "v:ic", ["verbose=", "ignore", "color"])

		for (option, value) in options:
			if option=="-v" or option=="--verbose":
				self.verbose = int(value)
			if option=="-i" or option=="--ignore":
				self.ignoreExceptions = 1
			if option=="-c" or option=="--color":
				self.color = 1
		self.build(*args)
		self.ignoreExceptions = ignoreExceptions
		self.verbose = verbose
		self.color = color

	def report(self, text):
		sys.stderr.write(text)

	def reportRule(self, text):
		if self.verbose>=1:
			self.report(text)

	def reportResult(self, text):
		if self.verbose>=2:
			self.report(text)

	def reportError(self, text):
		self.report(text)

class Dep:
	def __init__(self, source, target):
		self.source = source
		self.target = target

	def __repr__(self):
		return "<%s from %r to %r at 0x%x>" % (self.__class__.__name__, self.source, self.target, id(self))

class MainDep(Dep):
	pass

class Target:
	maxNewerReport = 2

	def __init__(self, project, id, action=None):
		self.project = project
		self.id = id
		self.action = action
		self._sources = []
		self._targets = []
		self.project.targets[self.id] = self

	def lastModified(self):
		"""
		return the timestamp, when this target was last modified.
		The timestamp will be cached, as os.path.getmtime() is quite
		slow on Windows and will be updated when a target is remade.
		"""
		try:
			return self._lastModified
		except AttributeError:
			self._lastModified = self._fetchLastModified()
			return self._lastModified

	def _fetchLastModified(self):
		"""
		return the timestamp of the last modification.
		Overwrite in subclasses.
		"""
		raise NotImplementedError()

	def touch(self):
		"""
		set the "last modified" timestamp to now.
		"""
		self._lastModified = DateTime.now()

	def update(self):
		t1 = DateTime.now()
		for dep in self._sources:
			dep.source.update()
		if self.hasToBeRebuilt():
			if self.action is None:
				raise ValueError("don't know how to build %s" % self)
			if self.lastModified() is None:
				reason = "(target doesn't exist)"
			else:
				newer = [ str(dep.source) for dep in self.newerSources() ]
				if len(newer):
					if len(newer)>self.maxNewerReport:
						reason = "(%d sources changed)" % len(newer)
					else:
						reason = ", ".join(newer)
				else:
					reason = ""
			msg = "%s <-%s--" % (str(self), self.project.actionStr(self.action))
			if reason:
				msg += " %s" % reason
			msg += " ...\n"
			self.project.reportRule(msg)
			failed = 0
			try:
				msg = self.action.execute(self)
			except (SystemExit, KeyboardInterrupt):
				raise
			except Exception, ex:
				if self.project.ignoreExceptions:
					failed = 1
				else:
					raise
			t2 = DateTime.now()
			if failed:
				import traceback
				traceback.print_exc()
				fullmsg = "%s%s after %s\n" % (indent, self.project.errorStr("failed"), self.project.secondsStr(t2-t1))
				self.project.reportError(fullmsg)
				self.project.failed += 1
			else:
				fullmsg = "%sdone after %s" % (indent, self.project.secondsStr(t2-t1))
				if msg is not None:
					fullmsg += "; %s" % msg
				fullmsg += "\n"
				self.project.reportResult(fullmsg)
				self.project.rebuilt += 1
			self.touch() # always touch the target, so building can continue even if this target failed

	def sources(self, type=Dep):
		"""
		return a list of all the source dependancies for <self/> that are a subtype
		of <pyref arg="type">type</pyref>.
		"""
		return [ dep for dep in self._sources if isinstance(dep, type) ]

	def targets(self, type=Dep):
		"""
		return a list of all the target dependancies for <self/> that are a subtype
		of <pyref arg="type">type</pyref>.
		"""
		return [ dep for dep in self._targets if isinstance(dep, type) ]

	def newerSources(self, type=Dep):
		"""
		return a list of all the sources that or newer the this one
		(or don't exist).
		"""
		selfLM = self.lastModified()
		sources = self.sources(type)
		if selfLM is None:
			return sources
		else:
			newer = []
			for dep in sources:
				sourceLM = dep.source.lastModified()
				if not sourceLM or sourceLM>selfLM:
					newer.append(dep)
			return newer

	def dependOn(self, *others):
		"""
		<doc:par>add one or more dependencies.</doc:par>

		<doc:par><pyref arg="others">others</pyref> may contain <pyref class="Target">Target</pyref>
		instances and subclasses of <pyref class="Dep">Dep</pyref>.</doc:par>
		"""
		dependencyClass = Dep
		for other in others:
			if isinstance(other, types.ClassType) and issubclass(other, Dep):
				dependencyClass = other
			else:
				dependency = dependencyClass(other, self)
				self._sources.append(dependency)
				other._targets.append(dependency)

	def hasToBeRebuilt(self):
		"""
		<doc:par>does <self/> have to be rebuilt (either because this target does not exist, or
		it is older than one of its sources)?</doc:par>
		"""
		selfLM = self.lastModified()
		if selfLM:
			return len(self.newerSources())>0
		else:
			return 1

	def __str__(self):
		try:
			name = self.__class__.name
		except AttributeError:
			name = self.__class__.__name__
		return "%s(%s)" % (name, self.project.idStr(self.id))

	def __repr__(self):
		return "<%s instance id=%r at 0x%x>" % (self.__class__.__name__, self.id, id(self))

class FileSystemTarget(Target):
	pass

class FileTarget(FileSystemTarget):
	name = "File"

	def _fetchLastModified(self):
		try:
			return self.id.getmtime()
		except OSError:
			return None

class DirectoryTarget(FileSystemTarget):
	name = "Dir"

	def _fetchLastModified(self):
		# only rebuild if it doesn't exist, if it exists do nothing
		try:
			timestamp = self.id.getmtime()
		except OSError:
			return None
		return DateTime.DateTime(1900, 1, 1)

	def hasToBeRebuilt(self):
		return self.lastModified() is None

class ImageTarget(FileTarget):
	name = "Image"

class JavascriptTarget(FileTarget):
	name = "JS"

class CascadingStyleSheetTarget(FileTarget):
	name = "CSS"

class HTMLTarget(FileTarget):
	name = "HTML"

	def __init__(self, project, id, action=None, base=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

class XMLTarget(FileTarget):
	name = "XML"

	def __init__(self, project, id, action=None, base=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

class DTDTarget(FileTarget):
	name = "DTD"

class PDFTarget(FileTarget):
	name = "PDF"

class PythonTarget(FileTarget):
	name = "Python"

	def __init__(self, project, id, action=None, fromlist=None):
		FileTarget.__init__(self, project, id, action)
		self.fromlist = fromlist

	def load(self):
		(dir, name) = os.path.split(str(self.id))
		(base, ext) = os.path.splitext(name)
		if base=="__init__":
			(dir, name) = os.path.split(dir)
			(base, ext) = os.path.splitext(name)
		sys.path.insert(0, dir)
		if self.fromlist is not None:
			base = ".".join(self.fromlist) + "." + base
		module = __import__(base, globals(), locals(), self.fromlist)
		del sys.path[0]
		return module

class XISTNamespaceTarget(PythonTarget):
	name = "XISTNS"

class XISTTarget(FileTarget):
	name = "XIST"

	def __init__(self, project, id, action=None, base=None):
		FileTarget.__init__(self, project, id, action)
		self.base = base

	def parse(self):
		from xist import parsers
		return parsers.parseFile(self.id, base=self.base)

class PhonyTarget(Target):
	name = "Phony"

	def __init__(self, project, id, action=None):
		if action is None:
			action = NoAction()
		Target.__init__(self, project, id, action)

	def _fetchLastModified(self):
		return None

class Action:
	"""
	<doc:par>An <pyref nolink>Action</pyref> is responsible for rebuilding a <pyref class="Target">Target</pyref>.</doc:par>

	<doc:par>An <pyref nolink>Action</pyref> is stateless and so may be used by multiple targets. Executing the
	action is done through a call to the method <pyref method="execute">execute</pyref>.</doc:par>
	"""
	def __init__(self):
		pass

	def execute(self, target):
		"""
		<doc:par>execute the action to rebuild <pyref arg="target">target</pyref> and
		return a one-line status/result message.</doc:par>
		"""
		raise NotImplementedError()

	def __str__(self):
		try:
			return self.__class__.name
		except AttributeError:
			return self.__class__.__name__

class CopyAction(Action):
	"""
	<doc:par>generate the target by copying the source.</doc:par>
	<doc:par>The <z>source</z> is the only <pyref class="Target">Target</pyref> on which
	the target depends through a <pyref class="MainDep">MainDep</pyref>.</doc:par>
	"""
	name = "Copy"

	def execute(self, target):
		deps = target.sources(MainDep)
		assert len(deps)==1, "copy must have one main source for target %r, has %r" % (target, deps)

		infile = deps[0].source.id.open("rb")
		data = infile.read()
		infile.close()
		outfile = target.id.open("wb")
		outfile.write(data)
		outfile.close()
		return "%s copied" % target.project.sizeStr(len(data))

class MkDirAction(Action):
	name = "MkDir"

	def execute(self, target):
		target.id.mkdir()

class XISTAction(Action):
	name = "XIST"

	def __init__(self, encoding=None, xhtml=None, mode=None, target=None, stage=None, lang=None, targetroot=None):
		self.encoding = encoding
		self.xhtml = xhtml
		self.mode = mode
		self.target = target
		self.stage = stage
		self.lang = lang
		self.targetroot = targetroot

	def loadNamespaces(self, target):
		namespaces = [ dep.source.load() for dep in target.sources() if isinstance(dep.source, XISTNamespaceTarget) ]
		return namespaces

	def converter(self):
		from xist import converters
		return converters.Converter(mode=self.mode, stage=self.stage, target=self.target, lang=self.lang)

	def publish(self, target, node):
		from xist import publishers
		outfile = target.id.open(mode="wb")
		publisher = publishers.FilePublisher(outfile, base=str(target.base), root=self.targetroot, encoding=self.encoding, xhtml=self.xhtml)
		node.publish(publisher)
		bytes = publisher.tell()
		outfile.close()
		return bytes

	def execute(self, target):
		self.loadNamespaces(target)
		xistdeps = target.sources(MainDep)
		assert len(xistdeps)==1, "XISTAction must have one main source for target %r, has %r" % (target, xistdeps)
		t1 = DateTime.now()
		node = xistdeps[0].source.parse()
		t2 = DateTime.now()
		converter = self.converter()
		node = node.convert(converter)
		t3 = DateTime.now()
		bytes = self.publish(target, node)
		t4 = DateTime.now()
		return "parse %s; convert %s; save %s; %s bytes generated" % (target.project.secondsStr(t2-t1), target.project.secondsStr(t3-t2), target.project.secondsStr(t4-t3), target.project.sizeStr(bytes))

	def __str__(self):
		options = []
		for optionname in ["encoding", "xhtml", "mode", "target", "stage", "lang"]:
			optionvalue = getattr(self, optionname)
			if optionvalue is not None:
				options.append("%s=%r" % (optionname, optionvalue))
		return "%s(%s)" % (Action.__str__(self), ", ".join(options))

class NoAction(Action):
	name = "None"

	def execute(self, target):
		pass

class CommandAction(Action):
	name = "Command"

	def __init__(self, command):
		Action.__init__(self)
		self.command = command

	def execute(self, target):
		os.system(self.command)

	def __str__(self):
		return "%s(%r)" % (Action.__str__(self), self.command)

